{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\nFormat osmFISH Data\n===================\n\nThe following script formats osmFISH data from primary human visual cortex in SpaceTx-Format.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import functools\nimport json\nimport os\nimport re\nfrom typing import Mapping, Union\n\nimport click\nimport numpy as np\nfrom slicedimage import ImageFormat\n\nimport starfish.core.util.try_import\nfrom starfish.experiment.builder import FetchedTile, TileFetcher, write_experiment_json\nfrom starfish.types import Axes, Coordinates, CoordinateValue, Features\n\n\n# We use this to cache images across tiles.  In the case of the osmFISH data set, volumes are saved\n# together in a single file.  To avoid reopening and decoding the TIFF file, we use a single-element\n# cache that maps between file_path and the npy file.\n@functools.lru_cache(maxsize=1)\ndef cached_read_fn(file_path) -> np.ndarray:\n    return np.load(file_path)\n\n\nclass osmFISHTile(FetchedTile):\n\n    def __init__(\n            self,\n            file_path: str,\n            coordinates: Mapping[Union[str, Coordinates], CoordinateValue],\n            z: int\n    ) -> None:\n        \"\"\"Parser for an osmFISH tile.\n\n        Parameters\n        ----------\n        file_path : str\n            location of the osmFISH tile\n        coordinates : Mapping[Union[str, Coordinates], CoordinateValue]\n            the coordinates for the selected osmFISH tile, extracted from the metadata\n        z : int\n            the z-layer for the selected osmFISH tile\n        \"\"\"\n        self.file_path = file_path\n        self.z = z\n        self._coordinates = coordinates\n\n    @property\n    def shape(self) -> Mapping[Axes, int]:\n        \"\"\"\n        Gets image shape directly from the data. Note that this will result in the data being\n        read twice, since the shape is retrieved from all tiles before the data is read, and thus\n        single-file caching does not resolve the duplicated reads.\n\n        Because the data here isn't tremendously large, this is acceptable in this instance.\n        \"\"\"\n        raw_shape = self.tile_data().shape\n        return {Axes.Y: raw_shape[0], Axes.X: raw_shape[1]}\n\n    @property\n    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:\n        return self._coordinates\n\n    def tile_data(self) -> np.ndarray:\n        return cached_read_fn(self.file_path)[self.z]  # slice out the correct z-plane\n\n\nclass osmFISHTileFetcher(TileFetcher):\n\n    @starfish.core.util.try_import.try_import({\"yaml\"})\n    def __init__(self, input_dir: str, metadata_yaml) -> None:\n        \"\"\"Implement a TileFetcher for an osmFISH experiment.\n\n        This TileFetcher constructs spaceTx format for one or more fields of view, where\n        `input_dir` is a directory containing all .npy image files and whose file names have the\n        following structure:\n\n        Hybridization<round>_<target>_fov_<fov_number>.npy\n\n        Notes\n        -----\n        - osmFISH is a non-multiplex method. As such, each target is specified by a\n          (channel, round) tuple. The files do not contain channel information,\n        - The spatial organization of the fields of view are not known to the starfish developers,\n          so they are filled by dummy coordinates\n        - This TileFetcher is specifically tailored to the gene panel used for a specific\n          experiment. Generalization of this TileFetcher will require reimplementation of the\n          `channel_map` method.\n        \"\"\"\n        import yaml\n\n        with open(metadata_yaml, \"r\") as f:\n            self.osmfish_metadata = yaml.load(f)\n        self.num_z = self.osmfish_metadata['ImageProperties']['HybImageSize']['zcount']\n        self.input_dir = input_dir\n\n    @property\n    def channel_map(self) -> Mapping[str, int]:\n        return {\n            \"Quasar570\": 0,\n            \"CF610\": 1,\n            \"Quasar670\": 2,\n        }\n\n    @property\n    def target_map(self) -> Mapping:\n        \"\"\"\n        Parse the metadata to map channel number and round to the correct target, which identifies\n        the right tile\n        \"\"\"\n        parsed_metadata = {}\n        for round_, round_data in self.osmfish_metadata['HybridizationsInfos'].items():\n\n            # the metadata references a larger corpus of data than we use for this example so as we\n            # iterate over the metadata, some rounds will not be found. In those cases, we just\n            # continue through the loop without adding to parsed_metadata\n            round_match = re.match(r\"Hybridization(\\d{1,2})\", round_)\n            if round_match is None:\n                continue\n\n            round_id = int(round_match.group(1)) - 1\n            for target_name, fluorophore in round_data.items():\n                if fluorophore in {\"Dapi\", \"FITC\"}:\n                    continue\n                channel = self.channel_map[fluorophore]\n                parsed_metadata[round_id, channel] = target_name\n\n        return parsed_metadata\n\n    @property\n    def fov_map(self) -> Mapping[int, str]:\n        \"\"\"This example dataset has three channels, which are mapped to sequential integers\"\"\"\n        return {\n            0: \"53\",\n            1: \"75\",\n            2: \"106\",\n        }\n\n    def coordinate_map(self, round_: int, z: int):\n        pixel_size = self.osmfish_metadata['ImageProperties']['PixelSize']\n        y_pixels, x_pixels, z_pixels = (\n            self.osmfish_metadata['ImageProperties']['HybImageSize'].values()\n        )\n        y_size = y_pixels * pixel_size\n        x_size = x_pixels * pixel_size\n        z_size = z_pixels * pixel_size\n        position_string = (\n            self.osmfish_metadata['TilesPositions'][f'Hybridization{round_ + 1}'][z]\n        )\n        y_pos, x_pos, z_pos = (float(v) for v in position_string.split(\", \"))\n        return {\n            Coordinates.X: (x_pos, x_pos + x_size),\n            Coordinates.Y: (y_pos, y_pos + y_size),\n            Coordinates.Z: (z_pos, z_pos + z_size),\n        }\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:\n        target = self.target_map[round_label, ch_label]\n        fov = self.fov_map[fov_id]\n        basename = f\"Hybridization{round_label + 1}_{target}_fov_{fov}.npy\"\n        file_path = os.path.join(self.input_dir, basename)\n        coordinates = self.coordinate_map(round_label, zplane_label)\n        return osmFISHTile(file_path, coordinates, zplane_label)\n\n    def generate_codebook(self):\n        mappings = []\n        for (round_, channel), target in self.target_map.items():\n            mappings.append({\n                Features.CODEWORD: [{\n                    Axes.ROUND.value: round_, Axes.CH.value: channel, Features.CODE_VALUE: 1\n                }],\n                Features.TARGET: target\n            })\n        return {\n            \"version\": \"0.0.0\",\n            \"mappings\": mappings\n        }\n\n\n@click.command()\n@click.argument(\"input-dir\", type=str)\n@click.argument(\"metadata-yaml\", type=str)\n@click.argument(\"output-dir\", type=str)\ndef cli(input_dir, metadata_yaml, output_dir):\n    \"\"\"Reads osmFISH images from <input-dir> and experiment metadata from <metadata-yaml> and writes\n    spaceTx-formatted data to <output-dir>.\n\n    Raw data (input for this tool) for this experiment can be found at:\n    s3://spacetx.starfish.data.upload/simone/\n\n    Processed data (output of this tool) can be found at:\n    s3://spacetx.starfish.data.public/20181031/osmFISH/\n    \"\"\"\n    os.makedirs(output_dir, exist_ok=True)\n    primary_tile_fetcher = osmFISHTileFetcher(os.path.expanduser(input_dir), metadata_yaml)\n\n    # This is hardcoded for this example data set\n    primary_image_dimensions = {\n        Axes.ROUND: 13,\n        Axes.CH: len(primary_tile_fetcher.channel_map),\n        Axes.ZPLANE: primary_tile_fetcher.num_z\n    }\n\n    def postprocess_func(experiment_json_doc):\n        experiment_json_doc[\"codebook\"] = \"codebook.json\"\n        return experiment_json_doc\n\n    with open(os.path.join(output_dir, \"codebook.json\"), \"w\") as f:\n        codebook = primary_tile_fetcher.generate_codebook()\n        json.dump(codebook, f)\n\n    write_experiment_json(\n        path=output_dir,\n        fov_count=len(primary_tile_fetcher.fov_map),\n        tile_format=ImageFormat.TIFF,\n        primary_image_dimensions=primary_image_dimensions,\n        aux_name_to_dimensions={},\n        primary_tile_fetcher=primary_tile_fetcher,\n        postprocess_func=postprocess_func,\n        dimension_order=(Axes.ROUND, Axes.CH, Axes.ZPLANE)\n    )\n    pass\n\n\nif __name__ == \"__main__\":\n    cli()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.8"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}